#include <algorithm>
#include <chrono>
#include <execution>
#include <iostream>
#include <ranges>
#include <string>

#include "matrix.h"

constexpr auto kPolicy = std::execution::par_unseq;

template <typename T>
void stepSlow(Matrix<T> &current, Matrix<T> &next) {
    auto rows = current.rows();
    auto cols = current.cols();

    auto idxs_rows = std::views::iota(1, rows - 1);
    std::for_each(kPolicy,
                  idxs_rows.begin(), idxs_rows.end(),
                  [cur = current.dataSpan(),
                   nxt = next.dataSpan(),
                   cols](int i) {
                      auto idxs_cols = std::views::iota(1, cols - 1);
                      std::for_each(idxs_cols.begin(), idxs_cols.end(),
                                    [cur, nxt, cols, row = i * cols](int j) {
                                        auto idx = row + j;  // we need a 1d index for the spans
                                        nxt[idx] = 0.25 * (cur[idx - 1] +
                                                           cur[idx + 1] +
                                                           cur[idx - cols] +
                                                           cur[idx + cols]);
                                    });
                  });
}

template <typename T>
void step(Matrix<T> &current, Matrix<T> &next) {
    auto cols = current.cols();
    auto rows = current.rows();

    auto up = -1 * cols;
    auto down = cols;
    auto left = -1;
    auto right = 1;

    auto idxs = std::views::iota(0, (cols - 2) * (rows - 2));

    std::for_each(kPolicy,
                  idxs.begin(), idxs.end(),
                  [cur = current.dataSpan(),
                   nxt = next.dataSpan(),
                   cols,
                   up, down, left, right](int idx) {
                      // We compute the 1d index to access the spans
                      auto r = idx / (cols - 2);
                      auto c = idx % (cols - 2);
                      auto i = (r + 1) * cols + c + 1;
                      nxt[i] = 0.25 * (cur[i + left] +
                                       cur[i + right] +
                                       cur[i + up] +
                                       cur[i + down]);
                  });
}

std::vector<int> computeIdxs(int rows, int cols) {
    std::vector<int> idxs((rows - 2) * (cols - 2), 0);

    auto iota = std::views::iota(0, (cols - 2) * (rows - 2));
    std::transform(kPolicy,
                   iota.begin(), iota.end(),
                   idxs.begin(),
                   [cols](int idx) {
                       // We compute the 1d index to access the spans
                       auto r = idx / (cols - 2);
                       auto c = idx % (cols - 2);
                       return (r + 1) * cols + c + 1;
                   });
    return idxs;
}

template <typename T>
void stepPreComputedIdxs(std::vector<int> &idxs, Matrix<T> &current, Matrix<T> &next) {
    auto up = -1 * current.cols();
    auto down = current.cols();
    auto left = -1;
    auto right = 1;

    std::for_each(kPolicy,
                  idxs.begin(), idxs.end(),
                  [cur = current.dataSpan(),
                   nxt = next.dataSpan(),
                   up, down, left, right](int i) {
                      nxt[i] = 0.25 * (cur[i + left] +
                                       cur[i + right] +
                                       cur[i + up] +
                                       cur[i + down]);
                  });
}

int main(int argc, char **argv) {
    int m = std::stoi(argv[1]);
    int n = std::stoi(argv[2]);
    int tmax = std::stoi(argv[3]);
    std::string filename = argv[4];

    auto start_global = std::chrono::steady_clock::now();

    std::cout << "processors : seq" << std::endl;
    std::cout << "m : " << m << std::endl;
    std::cout << "n : " << n << std::endl;
    std::cout << "tmax : " << tmax << std::endl;

    Matrix<double> *U = new Matrix<double>(m, n);
    Matrix<double> *Unext = new Matrix<double>(m, n);
    init_matrix(*U);
    init_matrix(*Unext);

    auto start_iter = std::chrono::steady_clock::now();
    // auto idxs = computeIdxs(U->rows(), U->cols());
    for (int t = 0; t < tmax; t++) {
        step(*U, *Unext);
        // stepPreComputedIdxs(idxs, *U, *Unext);
        std::swap(U, Unext);
        if (t % 100 == 0) {
            std::cout << "t:" << t << std::endl;
        }
    }
    auto end_iter = std::chrono::steady_clock::now();
    auto duration_iter = std::chrono::duration<double>(end_iter - start_iter);

    std::cout << "iterations time : " << duration_iter.count() << "[s]" << std::endl;

    write_mat(*U, filename);
    std::cout << "sum of elements of matrix : " << sum_mat(*U) << std::endl;

    delete U;
    delete Unext;

    auto end_global = std::chrono::steady_clock::now();
    auto duration_global = std::chrono::duration<double>(end_global - start_global);
    std::cout << "total time : " << duration_global.count() << "[s]" << std::endl;
}