#include <iostream>
#include <vector>

constexpr size_t kVecSize = 100000;

__device__ constexpr size_t getIdx() {
    return blockDim.x * blockIdx.x + threadIdx.x;
}

__global__ void sumVectors(float *a, float *b, float *c) {
    size_t idx = getIdx();
    if (idx < kVecSize) {
        c[idx] = a[idx] + b[idx];
    }
}

int main() {
    constexpr size_t kBytes = sizeof(float) * kVecSize;

    // Host initialization
    float *host_vec_a = nullptr;
    float *host_vec_b = nullptr;
    float *host_vec_c = nullptr;
    cudaMallocHost(&host_vec_a, kBytes);
    cudaMallocHost(&host_vec_b, kBytes);
    cudaMallocHost(&host_vec_c, kBytes);

    for (size_t i = 0; i < kVecSize; i++) {
        host_vec_a[i] = static_cast<float>(i);
        host_vec_b[i] = 2.0 * static_cast<float>(i);
    }

    // Device initialization
    float *dev_vec_a = nullptr;
    float *dev_vec_b = nullptr;
    float *dev_vec_c = nullptr;
    cudaMalloc(&dev_vec_a, kBytes);
    cudaMalloc(&dev_vec_b, kBytes);
    cudaMalloc(&dev_vec_c, kBytes);

    // Copy host vectors on device
    cudaMemcpy(dev_vec_a, host_vec_a, kBytes, cudaMemcpyHostToDevice);
    cudaMemcpy(dev_vec_b, host_vec_b, kBytes, cudaMemcpyHostToDevice);

    // Compute device execution parameters
    constexpr size_t kThreadPerBlock = 256;
    constexpr size_t kBlockPerGrid = kVecSize / kThreadPerBlock + ((kVecSize % kThreadPerBlock) != 0 ? 1 : 0);

    // Run kernel
    sumVectors<<<kBlockPerGrid, kThreadPerBlock>>>(dev_vec_a,
                                                   dev_vec_b,
                                                   dev_vec_c);

    // Get back the resulting vector to host
    cudaMemcpy(host_vec_c, dev_vec_c, kBytes, cudaMemcpyDeviceToHost);

    constexpr float tol = 1e-5;
    bool is_correct = true;
    for (size_t i = 0; i < kVecSize; i++) {
        float expected = 3.0 * static_cast<float>(i);
        if (std::abs(host_vec_c[i] - expected) > tol) {
            is_correct = false;
            break;
        }
    }
    std::cout << "The sum of the two vector is ";
    if (is_correct) {
        std::cout << "correct";
    } else {
        std::cout << "wrong";
    }
    std::cout << std::endl;

    // Host cleanup
    cudaFreeHost(host_vec_a);
    cudaFreeHost(host_vec_b);
    cudaFreeHost(host_vec_c);

    // Device cleanup
    cudaFree(dev_vec_a);
    cudaFree(dev_vec_b);
    cudaFree(dev_vec_c);
}
